﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.QuadraticAssignment.Algorithms {
  [Item("QAPRobustTabooSearchOperator", "Performs an iteration of the robust taboo search algorithm as descrbied in Taillard 1991.")]
  public sealed class QAPRobustTabooSeachOperator : SingleSuccessorOperator, IIterationBasedOperator, IStochasticOperator {

    #region Parameter Properties
    public ILookupParameter<IntValue> IterationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters["Iterations"]; }
    }
    public ILookupParameter<IntValue> TabuTenureParameter {
      get { return (ILookupParameter<IntValue>)Parameters["TabuTenure"]; }
    }
    public ILookupParameter<Permutation> PermutationParameter {
      get { return (ILookupParameter<Permutation>)Parameters["Permutation"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleMatrix> DistancesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public ILookupParameter<DoubleMatrix> ShortTermMemoryParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["ShortTermMemory"]; }
    }
    public ILookupParameter<DoubleMatrix> LongTermMemoryParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["LongTermMemory"]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<DoubleValue> BestQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["BestQuality"]; }
    }

    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }
    public IValueLookupParameter<IntValue> MaximumIterationsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumIterations"]; }
    }
    public IValueLookupParameter<IntValue> MinimumTabuTenureParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MinimumTabuTenure"]; }
    }
    public IValueLookupParameter<IntValue> MaximumTabuTenureParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumTabuTenure"]; }
    }
    public IValueLookupParameter<IntValue> TabuTenureAdaptionIntervalParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["TabuTenureAdaptionInterval"]; }
    }
    public IValueLookupParameter<BoolValue> UseAlternativeAspirationParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["UseAlternativeAspiration"]; }
    }
    public IValueLookupParameter<IntValue> AlternativeAspirationTenureParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["AlternativeAspirationTenure"]; }
    }
    #endregion

    [StorableConstructor]
    private QAPRobustTabooSeachOperator(bool deserializing) : base(deserializing) { }
    private QAPRobustTabooSeachOperator(QAPRobustTabooSeachOperator original, Cloner cloner)
      : base(original, cloner) {
    }
    public QAPRobustTabooSeachOperator() {
      Parameters.Add(new LookupParameter<IntValue>("Iterations", "The current iteration."));
      Parameters.Add(new LookupParameter<IntValue>("TabuTenure", "The current tabu tenure."));
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new LookupParameter<Permutation>("Permutation", "The permutation solution."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", "The weights matrix."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", "The distances matrix."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("ShortTermMemory", "The table that stores the iteration at which a certain facility has been assigned to a certain location."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("LongTermMemory", "Same as the tabu table, but constantly updates the information given the current solution."));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The quality value."));
      Parameters.Add(new LookupParameter<DoubleValue>("BestQuality", "The best quality value."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumIterations", "The number of iterations that the algorithm should run."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MinimumTabuTenure", "The minimum tabu tenure."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumTabuTenure", "The maximum tabu tenure."));
      Parameters.Add(new ValueLookupParameter<IntValue>("TabuTenureAdaptionInterval", "The amount of iterations that have to pass before the tabu tenure is adapted."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("UseAlternativeAspiration", "True if the alternative aspiration condition should be used that takes moves that have not been made for some time above others."));
      Parameters.Add(new ValueLookupParameter<IntValue>("AlternativeAspirationTenure", "The time t that a move will be remembered for the alternative aspiration condition."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new QAPRobustTabooSeachOperator(this, cloner);
    }

    public override IOperation Apply() {
      IRandom random = RandomParameter.ActualValue;
      int iteration = IterationsParameter.ActualValue.Value;
      if (iteration % TabuTenureAdaptionIntervalParameter.ActualValue.Value == 0)
        TabuTenureParameter.ActualValue.Value = MinimumTabuTenureParameter.ActualValue.Value
          + random.Next(MaximumTabuTenureParameter.ActualValue.Value - MinimumTabuTenureParameter.ActualValue.Value + 1);

      DoubleMatrix longTermMemory = LongTermMemoryParameter.ActualValue;
      DoubleMatrix shortTermMemory = ShortTermMemoryParameter.ActualValue;
      DoubleMatrix weights = WeightsParameter.ActualValue;
      DoubleMatrix distances = DistancesParameter.ActualValue;

      DoubleValue quality = QualityParameter.ActualValue;
      DoubleValue bestQuality = BestQualityParameter.ActualValue;
      if (bestQuality == null) {
        BestQualityParameter.ActualValue = (DoubleValue)quality.Clone();
        bestQuality = BestQualityParameter.ActualValue;
      }

      int tabuTenure = TabuTenureParameter.ActualValue.Value;
      int alternativeAspirationTenure = AlternativeAspirationTenureParameter.ActualValue.Value;
      Permutation solution = PermutationParameter.ActualValue;

      for (int i = 0; i < solution.Length; i++)
        longTermMemory[i, solution[i]] = iteration;

      double bestMoveQuality = double.MaxValue;
      Swap2Move bestMove = null;

      foreach (Swap2Move move in ExhaustiveSwap2MoveGenerator.Generate(solution)) {
        double moveQuality = QAPSwap2MoveEvaluator.Apply(solution, move, weights, distances);
        if (UseAlternativeAspirationParameter.ActualValue.Value
          && longTermMemory[move.Index1, solution[move.Index2]] < iteration - alternativeAspirationTenure
          && longTermMemory[move.Index2, solution[move.Index1]] < iteration - alternativeAspirationTenure) {
          bestMove = move;
          bestMoveQuality = moveQuality;
          break;
        }
        bool isTabu = shortTermMemory[move.Index1, solution[move.Index2]] > 0
          && shortTermMemory[move.Index2, solution[move.Index1]] > 0
          && ((iteration - shortTermMemory[move.Index1, solution[move.Index2]]) <= tabuTenure)
          && ((iteration - shortTermMemory[move.Index2, solution[move.Index1]]) <= tabuTenure);
        if (!isTabu || quality.Value + moveQuality < bestQuality.Value) {
          if (moveQuality < bestMoveQuality) {
            bestMoveQuality = moveQuality;
            bestMove = move;
          }
        }
      }

      if (bestMove == null) return base.Apply();

      shortTermMemory[bestMove.Index1, solution[bestMove.Index1]] = iteration;
      shortTermMemory[bestMove.Index2, solution[bestMove.Index2]] = iteration;
      Swap2Manipulator.Apply(solution, bestMove.Index1, bestMove.Index2);
      quality.Value += bestMoveQuality;

      if (quality.Value < bestQuality.Value) bestQuality.Value = quality.Value;

      return base.Apply();
    }
  }
}
